import json
import sys
import os

import logging

from trytond.transaction import Transaction

from pyairtable import Api
from marshmallow import Schema, fields, EXCLUDE, post_load
from marshmallow.exceptions import ValidationError
from bij1.erp.models import Member
from bij1.erp import pool


logger = logging.getLogger()
logger.setLevel(logging.INFO)
handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.INFO)
logger.addHandler(handler)

DEBUG = os.environ.get("DEBUG", False)


class MemberSchema(Schema):
    name = fields.Str()
    email = fields.Email()
    phone: fields.Str()
    address = fields.Str()
    postcode = fields.Str()
    city = fields.Str()
    status = fields.Str()
    since = fields.Date()
    until = fields.Date()
    iban = fields.Str()
    period = fields.Str()
    amount = fields.Decimal()
    mandate = fields.Str()

    class Meta:
        unknown = EXCLUDE

    @post_load
    def obj(self, data, **kwargs):
        return Member(**data)


class Airtable:
    field_map = {
        "Name": "name",
        "Straat": "address",
        "Postcode": "postcode",
        "Plaatsnaam": "city",
        "Lidmaatschapsstatus": "status",
        "Inschrijvingsdatum": "since",
        "Uitschrijvingsdatum": "until",
        "IBAN": "iban",
        "mollie_id": "mandate",
        "periode": "period",
        "termijnbedrag": "amount",
    }

    def __init__(self, key: str, base: str = "BIJ1", table: str = "leden"):
        self._api = Api(key)
        (base,) = [x for x in self._api.bases() if x.name == base]
        self._table = base.table(table)
        self._schema = MemberSchema()
        self.members: list[Member] = []

    def iterate(self):
        return self._table.iterate()

    def _dump(self, fd):
        records = []
        total = 0
        for page in self.iterate():
            for record in page:
                records.append(record)
                total += 1
                if total % 100 == 0:
                    logger.debug(" %s", total)
        json.dump(records, fd)

    def load(self, filename: str):
        if not os.path.exists(filename):
            with open(filename, "w") as fd:
                self._dump(fd)

        with open(filename, "r") as fd:
            records = json.load(fd)

        def clean(data: dict[str, str]):
            for old_key, new_key in self.field_map.items():
                if old_key in data:
                    data[new_key] = data[old_key]
                    del data[old_key]
            for key, value in data.items():
                if isinstance(value, str):
                    data[key] = value.strip()

        for record in records:
            data = record["fields"]
            clean(data)
            try:
                self.members.append(self._schema.load(data))
            except ValidationError as e:
                logger.info("skipping %s", e)

    def sync(self):
        logger.info("sync members...")
        with Transaction().start(pool.database_name, 0):
            total = len(self.members)
            for counter, member in enumerate(self.members):
                if counter % 100 == 0:
                    logger.info(f" {counter:>6}:{total}")
                if DEBUG and counter >= 1000:
                    break
                member.save()
        logger.info("sync members done")


def airtable():
    if not (key := os.environ.get("AIRTABLE_ACCESS_TOKEN", None)):
        logger.error("no AIRTABLE_ACCESS_TOKEN")
        return

    client = Airtable(key)
    client.load("airtable.json")
    client.sync()


if __name__ == "__main__":
    airtable()
