from dataclasses import dataclass
from schwifty import IBAN
from datetime import date
from decimal import Decimal
import typing
import logging
from bij1.erp import pool
from bij1.erp.config import config


logger = logging.getLogger(__name__)


@dataclass
class Member:
    name: str
    email: typing.Optional[str] = None
    phone: typing.Optional[str] = None
    address: typing.Optional[str] = None
    postcode: typing.Optional[str] = None
    city: typing.Optional[str] = None
    status: typing.Optional[str] = None
    since: typing.Optional[date] = None
    until: typing.Optional[date] = None
    iban: typing.Optional[str] = None
    period: typing.Optional[str] = None
    amount: typing.Optional[Decimal] = None
    mandate: typing.Optional[str] = None

    def noop(self):
        logger.debug("noop called for %s", self)

    def exists(self):
        ContactMechanism = pool.get("party.contact_mechanism")
        result = ContactMechanism.search(
            [
                ("type", "=", "email"),
                ("value", "=", self.email),
                ("active", "=", True),
            ]
        )
        return bool(result)

    def save_party(self):
        Party = pool.get("party.party")
        if not (obj := Party.search([("name", "=", self.name)])):
            obj = Party()
            obj.name = self.name
            obj.save()
        else:
            obj = obj[0]
        self.party = obj

    def save_address(self):
        Address = pool.get("party.address")
        if not (address := self.party.address_get()):
            address = Address()
            address.party = self.party.id

        address.street = self.address
        address.postal_code = self.postcode
        address.city = self.city
        address.invoice = True
        address.save()

    def save_email(self):
        ContactMechanism = pool.get("party.contact_mechanism")
        if not ContactMechanism.search(
            [
                ("party", "=", self.party.id),
                ("type", "=", "email"),
                ("value", "=", self.email),
                ("active", "=", True),
            ]
        ):
            contact_mechanism = ContactMechanism()
            contact_mechanism.party = self.party.id
            contact_mechanism.type = "email"
            contact_mechanism.active = True
            contact_mechanism.value = self.email
            contact_mechanism.save()

    def clean_ibans(self) -> list[str]:
        result = self.iban.split(",")
        result = [x.replace(" ", "") for x in result]
        return [x.upper() for x in result]

    def save_iban(self):
        if not self.iban:
            return

        Party = pool.get("party.party")
        Bank = pool.get("bank")
        BankAccount = pool.get("bank.account")
        BankAccountNumber = pool.get("bank.account.number")
        for iban in self.clean_ibans():
            try:
                parsed = IBAN(iban)
                bank_name = parsed.bank_code
            except Exception:
                logger.error("invalid iban: %s", iban)
                continue

            if not (bank := Bank.search([("party", "=", bank_name)])):
                if not (bank_party := Party.search([("name", "=", bank_name)])):
                    bank_party = Party()
                    bank_party.name = bank_name
                    bank_party.save()
                else:
                    bank_party = bank_party[0]

                bank = Bank()
                bank.party = bank_party
            else:
                bank = bank[0]

            bank.bic = parsed.bic
            bank.save()

            if not (
                bank_account := BankAccount.search(
                    [
                        ("owners", "in", [self.party.id]),
                        ("bank", "=", bank),
                        ("numbers.number", "in", [iban]),
                    ]
                )
            ):
                bank_account = BankAccount()
                bank_account.bank = bank
                bank_account.owners = [self.party.id]
                bank_account.save()
            else:
                bank_account = bank_account[0]

            if not (
                account_number := BankAccountNumber.search(
                    [
                        ("account", "=", bank_account.id),
                        ("number", "=", iban),
                    ]
                )
            ):
                account_number = BankAccountNumber()
                account_number.account = bank_account
                account_number.number = iban
                account_number.save()

    def save_mandate(self):
        if not (self.iban and self.mandate):
            return
        Mandate = pool.get("account.payment.sepa.mandate")
        if Mandate.search([("identification", "=", self.mandate)]):
            return

        BankAccount = pool.get("bank.account")
        bank_account = None
        for iban in self.clean_ibans():
            if bank_account := BankAccount.search(
                [("owners", "in", [self.party.id]), ("numbers.number", "in", [iban])]
            ):
                # use the first account for now
                bank_account = bank_account[0]
                break
        if not bank_account:
            return
        account_number = bank_account.numbers[0]
        mandate = Mandate()
        mandate.identification = self.mandate
        mandate.company = config.company
        mandate.account_number = account_number
        mandate.party = self.party
        mandate.signature_date = self.since
        mandate.save()

    def save_subscription(self):
        if not self.period:
            return

        Subscription = pool.get("sale.subscription")
        SubscriptionLine = pool.get("sale.subscription.line")
        SubscriptionService = pool.get("sale.subscription.service")
        services = SubscriptionService.search([("active", "=", True)])
        services = {x.consumption_recurrence.name: x for x in services}

        if not (service := services.get(self.period.lower())):
            raise Exception("unkown period: %s" % self.period)

        if not (subscription := Subscription.search(["party", "=", self.party])):
            subscription = Subscription()
            subscription.company = config.company
            subscription.party = self.party
            subscription.currency = config.currency
            subscription.start_date = self.since
            subscription.invoice_start_date = date(2024, 1, 1)
            subscription.invoice_address = self.party.addresses[0]
            subscription.invoice_recurrence = service.consumption_recurrence
        else:
            subscription = subscription[0]

        if self.until:
            subscription.end_date = self.until

        subscription.description = config.subscription_description

        if self.until and self.until < date(2024, 1, 1):
            if lines := SubscriptionLine.search(
                [("subscription", "=", subscription), ("service", "=", service)]
            ):
                SubscriptionLine.delete(lines)
            return

        subscription.save()

        if not (
            line := SubscriptionLine.search(
                [("subscription", "=", subscription), ("service", "=", service)]
            )
        ):
            line = SubscriptionLine()
            line.service = service
            line.subscription = subscription
            line.quantity = 1
            line.unit_price = self.amount
            line.start_date = max([self.since, date(2024, 1, 1)])
            line.end_date = self.until
            line.unit = service.product.default_uom
        else:
            line = line[0]

        line.save()
        Subscription.quote([subscription])
        Subscription.run([subscription])

    def update(self):
        # XXX: to be improved
        self.save()

    def create(self):
        # XXX: to be improved
        self.save()

    def save(self):
        self.save_party()
        self.save_address()
        self.save_email()
        self.save_iban()
        self.save_mandate()
        self.save_subscription()
        self.party.save()
